########################################################################
# gdipcgi.pm
#
# This file really is the GnuDIP CGI. The gnudip.cgi script just
# executes the gdiprun subroutine.
#
# See COPYING for licensing information.
#
# Written by:
#
#   Creighton MacDonnell
#   http://macdonnell.ca/
#
########################################################################

# Perl modules
use strict;
use CGI;

# global variables
use vars qw($conf $gnudipdir);
use vars qw($pref $userinfo $dominfo $remote_ip);
use vars qw($reqparm $thiscgi $thishost $logger);

# GnuDIP common subroutines
use gdiplib;
use dbprefs;
use dbusers;
use doreq;
use htmlgen;

########################################################################
# NOTE
# Each pg_? routine generates a page and does an "exit".
# So they DO NOT RETURN.
########################################################################

########################################################################
# NOTE
# The do_? routines all call a pg_? routine.
# So they DO NOT RETURN.
########################################################################

########################################################################
# the actual CGI
########################################################################
sub gdiprun {

  # get preferences from config file
  $conf = getconf();
  if (! $conf) {
    print STDERR "GnuDIP CGI has exited - getconf returned nothing\n";
    htmlgen_init();
    dberror();
  }

  # logger command
  $logger = $$conf{'logger_cgi'};

  # not called by HTTP server
  if (!$ENV{'REQUEST_METHOD'} or
      $ENV{'REQUEST_METHOD'} ne 'GET' and
      $ENV{'REQUEST_METHOD'} ne 'POST') {
    writelog( 'gdipcgi.pm: REQUEST_METHOD environment variable is not set');
    print "REQUEST_METHOD environment variable is not set\n";
    writelog( 'gdipcgi.pm: Not called by HTTP server');
    print "Not called by HTTP server\n";
    exit;
  }

  # no current user to start
  $userinfo  = undef;
  $dominfo   = undef;

  # IP address of connecting machine
  $remote_ip   = $ENV{'REMOTE_ADDR'};
  $remote_ip = '0.0.0.0' if !$remote_ip;

  # CGI request information
  my $query = new CGI;
  $thiscgi  = $query->url();
  $thishost = $query->virtual_host();
  my %reqparmhash;
  $reqparm  = \%reqparmhash;
  foreach my $name ($query->param()) {
    $$reqparm{$name} = join("\0", $query->param($name));
  }
  
  # log the HTTP request info to HTTP server log 
  # - for debugging
  #logreq();

  # initialise HTML generation routines
  htmlgen_init();

  # get preferences from the database
  $pref = getprefs();

  #############################################################
  #  identify requests not needing user validation
  #############################################################

  # deal with GET requests
  if ($ENV{'REQUEST_METHOD'} eq 'GET') {

    # main login page
    pg_login() if !$ENV{'QUERY_STRING'};

    # ensure initialized
    default_empty('login');
    default_empty('action');

    # quick login
    do_login() if $$reqparm{'login'} eq 'enc';

    # IP detection before quick login
    pg_ipdetect() if $$reqparm{'login'} eq 'ipdetect';

    # self registration
    pg_doself() if $$reqparm{'action'} eq 'signup';

    # do AutoURL
    do_autourl() if $$reqparm{'action'} eq 'getautourlinfo';

    # self registration after E-mail sent
    do_selfcreate() if $$reqparm{'selfcreate'};

    # E-mail update after E-mail sent
    do_newemail() if $$reqparm{'newemail'};

    # bad HTTP request
    pg_error('bad_request');
  }

  #############################################################
  #  only POST method from a GnuDIP page valid from here on
  #############################################################

  pg_error('bad_request') if $ENV{'REQUEST_METHOD'} ne 'POST';
  pg_error('bad_request') if !$$reqparm{'page'};

  # from login page?
  if ($$reqparm{'page'} eq 'login') {

    # login
    do_login() if $$reqparm{'login'};

    # forgotten password
    pg_sendURL() if $$reqparm{'sendURL'};

    # self register
    pg_doself() if $$reqparm{'self_signup'};

    pg_error('bad_request');
  }

  # from forgotten password page?
  if ($$reqparm{'page'} eq 'sendURL') {

    # send Quick login URL
    do_sendURL() if $$reqparm{'do_sendURL'};

    # pressed enter - reshow
    pg_sendURL() if defined $$reqparm{'sendURL_username'};

    pg_error('bad_request');
  }

  # from self registration page?
  if ($$reqparm{'page'} eq 'doself') {

    # self register
    do_self() if $$reqparm{'do_signup'};

    pg_error('bad_request');
  }

  #############################################################
  #  user validation needed from here on
  #############################################################

  # these form fields are created by the html_user subroutine
  # in htmlgen.pl
  default_empty('username');
  default_empty('domain');
  default_empty('password');
  default_empty('logonid');
  default_empty('pagetime');
  default_empty('checkval');

  # ensure people come through login page
  # validate the signature
  $$pref{'SERVER_KEY'} = '' if !defined($$pref{'SERVER_KEY'});
  my $checkval  =
    md5_hex(
      "$$reqparm{'username'}." .
      "$$reqparm{'domain'}." .
      "$$reqparm{'password'}." .
      "$$reqparm{'logonid'}." .
      "$$reqparm{'pagetime'}." .
      $$pref{'SERVER_KEY'}
      );
  pg_error('bad_request') if $checkval ne $$reqparm{'checkval'};

  # validate/retrieve the user
  $userinfo = getuser($$reqparm{'username'}, $$reqparm{'domain'});
  pg_error('nouser')
    if !$userinfo or $$userinfo{'username'} ne $$reqparm{'username'};
  pg_error('badpass') if $$userinfo{'password'} ne $$reqparm{'password'};

  # page expired?
  pg_error('page_timeout')
    if $$userinfo{'level'} ne 'ADMIN' and
       $$pref{'PAGE_TIMEOUT'}         and 
       $$reqparm{'pagetime'} + $$pref{'PAGE_TIMEOUT'} < time;

  # get information for user's domain
  if ($$reqparm{'domain'}) {
    $dominfo = getdomain($$reqparm{'domain'});
    pg_error('unknown_dom') if !$dominfo;
  }

  #############################################################
  #  identify requests needing user validation
  #############################################################

  # from E-mail needed page?
  if ($$reqparm{'page'} eq 'needemail') {

    # send E-mail update E-mail
    do_needemail() if $$reqparm{'do_needemail'};

    # pressed enter - reshow
    pg_needemail() if defined $$reqparm{'new_needemail'};

    pg_error('bad_request');
  }

  # from options page?
  if ($$reqparm{'page'} eq 'options') {

    #############################
    # common to ADMIN and USER
    #############################

    # user settings page
    pg_usersettings() if $$reqparm{'changesettings'};

    # set Quick Login URL page
    pg_setquick() if $$reqparm{'setquick'};

    # delete current user
    do_delthisuser() if $$reqparm{'do_delthisuser'};

    #############################
    # for USER
    #############################

    # update IP address
    do_updatehost() if $$reqparm{'updatehost'};

    # offline
    do_offline() if $$reqparm{'offline'};

    # user settings page
    pg_usersettings() if $$reqparm{'changesettings'};

    # set auto URL page
    pg_setautourl() if $$reqparm{'setautourl'};

    # pressed enter - reshow
    pg_options() if defined $$reqparm{'updateaddr'};

    #############################
    # for ADMIN
    #############################

    # add user page
    pg_adduser() if $$reqparm{'manageusers_adduser'};

    # manage users page
    do_manageusers() if $$reqparm{'manageusers_main'};

    # add domain page
    pg_adddomain() if $$reqparm{'managedomains_add'};

    # manage domains page
    pg_managedomains() if $$reqparm{'managedomains_main'};

    # systems settings page
    pg_syssettings() if $$reqparm{'system_settings'};

    # pressed enter - reshow
    pg_options() if defined $$reqparm{'user_pattern'};

    pg_error('bad_request');
  }

  # from user settings page?
  if ($$reqparm{'page'} eq 'updatesettings') {

    # update current user
    do_updatesettings() if $$reqparm{'do_updatesettings'};

    pg_error('bad_request');
  }

  # from set Auto URL page?
  if ($$reqparm{'page'} eq 'setautourl') {

    # test
    do_autourl() if $$reqparm{'testautourl'};

    # remove
    do_removeautourl() if $$reqparm{'removeautourl'};

    pg_error('bad_request');
  }

  # from add user page?
  if ($$reqparm{'page'} eq 'adduser') {

    # add user
    do_adduser() if $$reqparm{'do_adduser'};

    pg_error('bad_request');
  }

  # from manage users page?
  if ($$reqparm{'page'} eq 'manageusers') {

    # edit user page
    pg_edituser() if $$reqparm{'manageusers_edituser'};

    # delete users
    do_deluser() if $$reqparm{'do_deluser'};

    # reorder page contents
    do_manageusers()
      if $$reqparm{'manage_users_sortby_username'}  or
         $$reqparm{'manage_users_sortby_currentip'} or
         $$reqparm{'manage_users_sortby_domain'}    or
         $$reqparm{'manage_users_sortby_updated'}   or
         $$reqparm{'manage_users_sortby_level'}
         ;

    pg_error('bad_request');
  }

  # from user edit (by admin) page?
  if ($$reqparm{'page'} eq 'edituser') {

    # update user
    do_edituser() if $$reqparm{'do_edituser'};

    pg_error('bad_request');
  }

  # from add domain page?
  if ($$reqparm{'page'} eq 'adddomain') {

    # add domain
    do_adddomain() if $$reqparm{'do_adddomain'};

    # pressed enter - reshow
    pg_adddomain() if defined $$reqparm{'adddomain_new_domain'};

    pg_error('bad_request');
  }

  # from manage domains page?
  if ($$reqparm{'page'} eq 'managedomains') {

    # edit domain page
    pg_editdomain() if $$reqparm{'managedomains_edit'};

    # delete domains
    do_deldomain() if $$reqparm{'do_deldomain'};

    pg_error('bad_request');
  }

  # from edit domain page?
  if ($$reqparm{'page'} eq 'editdomain') {

    # update domain
    do_editdomain() if $$reqparm{'do_editdomain'};

    pg_error('bad_request');
  }

  # from system settings page?
  if ($$reqparm{'page'} eq 'syssettings') {

    # save system settings
    do_syssettings() if $$reqparm{'save_settings'};

    pg_error('bad_request');
  }

  #############################################################
  # bad HTTP request
  #############################################################
  pg_error('bad_request');

}

#############################################################
# common subroutines
#############################################################

# write to the log
sub writelog {
  calllogger($logger, @_) or pg_error('config');
}

# called for database error
sub dberror {
  pg_error('config');
}

# initialise request parameters
sub default_NO {
  my $key = shift;
  $$reqparm{$key} = 'NO'
    if !defined($$reqparm{$key}) or
                $$reqparm{$key} ne 'YES';
}
sub default_empty {
  my $key = shift;
  $$reqparm{$key} = '' if !defined($$reqparm{$key});
}

# URI escape a string
sub uri_escape
{
  my $text = shift;
  $text = '' if !defined($text);

  # set up escape values
  my %escapes;
  set_escape(\%escapes, ';');
  set_escape(\%escapes, '/');
  set_escape(\%escapes, '?');
  set_escape(\%escapes, ':');
  set_escape(\%escapes, '@');
  set_escape(\%escapes, '=');
  set_escape(\%escapes, '&');
  set_escape(\%escapes, '<');
  set_escape(\%escapes, '>');
  set_escape(\%escapes, '"');
  set_escape(\%escapes, '#');
  set_escape(\%escapes, '%');
  set_escape(\%escapes, '{');
  set_escape(\%escapes, '}');
  set_escape(\%escapes, '|');
  set_escape(\%escapes, '\\');
  set_escape(\%escapes, '^');
  set_escape(\%escapes, '~');
  set_escape(\%escapes, '[');
  set_escape(\%escapes, ']');
  set_escape(\%escapes, '`');
  set_escape(\%escapes, '+');

  # map unsafe characters (RFC 2732)
  $text =~ s/([\;\/\?\:\@\=\&\<\>\"\#\%\{\}\|\\\^\~\[\]\`\+])/$escapes{$1}/g;

  return $text;
}
sub set_escape {
  my $e = shift;
  my $c = shift;
  $$e{$c} = sprintf("%%%02X", ord($c));
}

# display the CGI data in the HTTP server log
sub logreq {
  my $var;
  my $val;
  print STDERR "ENV:\n";
  foreach $var (sort(keys(%ENV))) {
    $val = $ENV{$var};
    $val =~ s|\n|\\n|g;
    $val =~ s|"|\\"|g;
    print STDERR "  ${var}=\"${val}\"\n";
  }
  print STDERR "reqparm:\n";
  foreach $var (sort(keys(%$reqparm))) {
    $val = $$reqparm{$var};
    $val =~ s|\n|\\n|g;
    $val =~ s|"|\\"|g;
    print STDERR "  ${var}=\"${val}\"\n";
  }
}

#####################################################
# must return 1
#####################################################
1;

